#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/cdev.h>
#include <linux/slab.h>
#include <asm/uaccess.h>
#include <linux/usb.h>
#include <linux/string.h>

#include <linux/uaccess.h>
#include <linux/compat.h>
#include "rds-01-public.h"

#include "rds-01.h"
#include "touchmdef.h"

#define SMALL_BUF_SIZE 128
#define SINLE_ALLOC    1

#define to_devctx(d) container_of(d, struct _rds01_ctx, kref)

static __u32       DevCnt;
static PRDS_01_CTX *pDevList;
static struct class *touchm_drv_class;
static char *pDumpBuffer;

PRDS_01_CTX PrivFindDevCtx(int Minor)
{
  int i;
  for(i=0;i<TOUCHM_MAX_DEV_CNT;i++)
  {
    if(pDevList[i]==NULL)
      continue;
    if(pDevList[i]->Minor==Minor)
      return pDevList[i];
  }
  return NULL;
}


int priv_dev_probe(struct usb_interface *interface,
                 struct cdev *pCharDev,int Minor)
{
  int ret,i;
  PRDS_01_CTX pNewDev;
  struct usb_host_interface *iface_desc;
  struct usb_endpoint_descriptor *pEP;
  pNewDev=(PRDS_01_CTX)kzalloc(sizeof(RDS_01_CTX),GFP_KERNEL);
  memset(pNewDev,0,sizeof(RDS_01_CTX));
  pNewDev->Minor=Minor+TOUCHM_MINOR_BASE;
  pNewDev->pCharDev=pCharDev;
  pNewDev->pClassDev=device_create(touchm_drv_class,NULL,MKDEV(TOUCHM_MAJOR,Minor+TOUCHM_MINOR_BASE),NULL,"touchm%d",Minor);
  pNewDev->interface=interface;
  pNewDev->udev=usb_get_dev(interface_to_usbdev(interface));
  pDevList[Minor]=pNewDev;
  usb_set_intfdata(interface,pNewDev);
  DebugPrint(KERN_INFO "add new rds-01 attached to /dev/touchm%d\n",Minor);
	iface_desc = interface->cur_altsetting;  
  for(i=0;i<iface_desc->desc.bNumEndpoints;i++)
  {
    pEP=&iface_desc->endpoint[i].desc;
    if(usb_endpoint_is_bulk_in(pEP))
    {
      pNewDev->bIn_Addr=pEP->bEndpointAddress;
      pNewDev->bIn_size=le16_to_cpu(pEP->wMaxPacketSize);  
    }
    if(usb_endpoint_is_bulk_out(pEP))
    {
      pNewDev->bOut_Addr=pEP->bEndpointAddress;
      pNewDev->bOut_size=le16_to_cpu(pEP->wMaxPacketSize);  
    }
  }
  if(pNewDev->bIn_Addr*pNewDev->bOut_Addr==0)
  {
    DebugPrint(KERN_ERR "can not find in or out bulk endpoints\n");
    return -ENOMEM;
  }
  DebugPrint(KERN_INFO "in ep=>%x %x out ep=>%x %x\n",
        pNewDev->bIn_Addr,pNewDev->bIn_size,pNewDev->bOut_Addr,pNewDev->bOut_size);
#ifndef SINLE_ALLOC
  pNewDev->pInBuffer=kzalloc(pNewDev->bIn_size,GFP_DMA);
  pNewDev->pOutBuffer=kzalloc(pNewDev->bOut_size,GFP_DMA);
  pNewDev->pCtrlBuffer=kzalloc(64,GFP_DMA);
  pNewDev->pTXWrkBuffer=kzalloc(128,GFP_DMA);
  pNewDev->pRXWrkBuffer=kzalloc(128,GFP_DMA);
  pNewDev->pTXBlockBuffer=kzalloc(128,GFP_DMA);
  pNewDev->pRXBlockBuffer=kzalloc(128,GFP_DMA);
  pNewDev->pWriteBuffer=kzalloc(128,GFP_DMA);
  pNewDev->pDMACtrlBuf=kzalloc(64,GFP_DMA);
  pNewDev->pTmpRXBuf=kzalloc(64,GFP_DMA);
#else
  pNewDev->pTXWrkBuffer=kzalloc(4096,GFP_DMA);
  pNewDev->pRXWrkBuffer=&pNewDev->pTXWrkBuffer[SMALL_BUF_SIZE];
  pNewDev->pTXBlockBuffer=&pNewDev->pTXWrkBuffer[SMALL_BUF_SIZE*2];
  pNewDev->pRXBlockBuffer=&pNewDev->pTXWrkBuffer[SMALL_BUF_SIZE*3];
  pNewDev->pWriteBuffer=&pNewDev->pTXWrkBuffer[SMALL_BUF_SIZE*4];
  pNewDev->pDMACtrlBuf=&pNewDev->pTXWrkBuffer[SMALL_BUF_SIZE*5];
  pNewDev->pTmpRXBuf=&pNewDev->pTXWrkBuffer[SMALL_BUF_SIZE*6];
  pNewDev->pCtrlBuffer=&pNewDev->pTXWrkBuffer[SMALL_BUF_SIZE*7];
  pNewDev->pInBuffer=&pNewDev->pTXWrkBuffer[SMALL_BUF_SIZE*8];
  pNewDev->pOutBuffer=&pNewDev->pTXWrkBuffer[SMALL_BUF_SIZE*9];
  pNewDev->pTXUrb=usb_alloc_urb(0, GFP_KERNEL);
  pNewDev->pRXUrb=usb_alloc_urb(0, GFP_KERNEL);  
#endif  
  pNewDev->IsRemoved=0;
  //init_waitqueue_head(&pNewDev->IOPendingQ);
  kref_init(&pNewDev->kref);
  mdelay(20);
  pl2303_init(pNewDev);
  pl2303_set_speed(pNewDev,115200);
  return 0;
}

void priv_dev_delete(struct kref *kref)
{
  PRDS_01_CTX pDev;
  pDev=to_devctx(kref);
  DebugPrint(KERN_INFO "delete rds-01 /dev/touchm%d\n",pDev->Minor-TOUCHM_MINOR_BASE);
  cdev_del(pDev->pCharDev);
  device_destroy(touchm_drv_class,MKDEV(TOUCHM_MAJOR,pDev->Minor));
  pDevList[pDev->Minor-TOUCHM_MINOR_BASE]=NULL;
  pDev->IsRemoved=1;
  usb_free_urb(pDev->pTXUrb);
  usb_free_urb(pDev->pRXUrb);
#ifndef SINLE_ALLOC  
  kfree(pDev->pDMACtrlBuf);
  kfree(pDev->pTmpRXBuf);
  kfree(pDev->pInBuffer);
  kfree(pDev->pOutBuffer);
  kfree(pDev->pCtrlBuffer);
  kfree(pDev->pTXWrkBuffer);
  kfree(pDev->pRXWrkBuffer);
  kfree(pDev->pTXBlockBuffer);
  kfree(pDev->pRXBlockBuffer);
  kfree(pDev->pWriteBuffer);
  kfree(pDev->pTXBlockBuffer);
  kfree(pDev->pRXBlockBuffer);
#else
  kfree(pDev->pTXWrkBuffer);
#endif
  kfree(pDev->pCharDev);
  kfree(pDev);
}

int priv_dev_disconnect(struct usb_interface *interface)
{
  PRDS_01_CTX pDev;
  pDev=(PRDS_01_CTX)usb_get_intfdata(interface);
  usb_set_intfdata(interface,NULL);
  DebugPrint(KERN_INFO "disconnect rds-01\n");
  kref_put(&pDev->kref,priv_dev_delete);  
  return 0;
}

int priv_drv_init()
{
  pDumpBuffer=kzalloc(1024,GFP_KERNEL);
  if(register_chrdev_region(MKDEV(TOUCHM_MAJOR,TOUCHM_MINOR_BASE),TOUCHM_MAX_DEV_CNT,"rds-01")<0)
  {
    DebugPrint(KERN_ERR "touchm: Can not register char region.\n");
    return -1;
  }
  touchm_drv_class=class_create(THIS_MODULE,"touchm");
  DevCnt=0;
  pDevList=(PRDS_01_CTX*)kzalloc(sizeof(PRDS_01_CTX)*TOUCHM_MAX_DEV_CNT,GFP_KERNEL);
  memset(pDevList,0,sizeof(PRDS_01_CTX)*TOUCHM_MAX_DEV_CNT);
  return 0;
}

int priv_drv_close()
{
  kfree(pDevList);  
  unregister_chrdev_region(MKDEV(TOUCHM_MAJOR,TOUCHM_MINOR_BASE),TOUCHM_MAX_DEV_CNT);
  class_destroy(touchm_drv_class);
  kfree(pDumpBuffer);
  return 0;
}

int priv_drv_get_minor()
{
  int i;
  for(i=0;i<TOUCHM_MAX_DEV_CNT;i++)
  {
    if(pDevList[i]==NULL)
      return i;
  }
  return -1;
}
