#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <asm/uaccess.h>
#include <linux/usb.h>
#include <linux/string.h>

#include "touchmdef.h"
#include "rds-01.h"
#include "rds-01-public.h"
#define GET_LINE_REQUEST_TYPE	  0xa1
#define GET_LINE_REQUEST          0x21
#define SET_LINE_REQUEST_TYPE     0x21
#define SET_LINE_REQUEST          0x20
#define VENDOR_WRITE_REQUEST_TYPE 0x40
#define VENDOR_WRITE_REQUEST      0x01

#define VENDOR_READ_REQUEST_TYPE  0xc0
#define VENDOR_READ_REQUEST       0x01

enum pl2303_type 
{
  type_0,		/* don't know the difference between type 0 and */
  type_1,		/* type 1, until someone from prolific tells us... */
  HX,		/* HX version of the pl2303 chip */
};

static int pl2303_vendor_read(u16 value, u16 index,PRDS_01_CTX pDev,u8 *buf)
{
  int res=usb_control_msg(pDev->udev, usb_rcvctrlpipe(pDev->udev,0),
                          VENDOR_READ_REQUEST, VENDOR_READ_REQUEST_TYPE,
                         value,index,buf,1,100);
  if(res<0)
  {
    DebugPrint(KERN_INFO "vendor_read failed %d\n",res);
  }                         
#if 0                         
  DebugPrint(KERN_INFO "0x%x:0x%x:0x%x:0x%x  %d - %x", VENDOR_READ_REQUEST_TYPE,
         VENDOR_READ_REQUEST, value, index, res, buf[0]);
#endif         
  return res;
}

static int pl2303_vendor_write(u16 value, u16 index,PRDS_01_CTX pDev)
{
  int res = usb_control_msg(pDev->udev, usb_sndctrlpipe(pDev->udev,0),
			VENDOR_WRITE_REQUEST, VENDOR_WRITE_REQUEST_TYPE,
			value, index, NULL, 0, 100);
  if(res<0)
  {
    DebugPrint(KERN_INFO "vendor_write failed %d\n",res);
  }
#if 0			
	DebugPrint(KERN_INFO "0x%x:0x%x:0x%x:0x%x  %d", VENDOR_WRITE_REQUEST_TYPE,
			VENDOR_WRITE_REQUEST, value, index, res);
#endif			
	return res;
}


static int pl2303_get_line_request(PRDS_01_CTX pDev,u8 *buf)
{
  int ret;
  ret = usb_control_msg(pDev->udev,usb_rcvctrlpipe(pDev->udev,0),
                        GET_LINE_REQUEST, GET_LINE_REQUEST_TYPE,
				                0, 0, buf, 7, 100);
  if(ret!=7)
  {
    DebugPrint(KERN_INFO "get line reqw failed %d\n",ret);
    return -EIO;
  }
  //dumpData("regw=>",buf,7);
  return 0;
}

static int pl2303_set_line_request(PRDS_01_CTX pDev,u8 *buf)
{
  int ret;
  //dumpData("regw=>",buf,7);
   ret=usb_control_msg(pDev->udev,usb_sndctrlpipe(pDev->udev,0),
                       SET_LINE_REQUEST, SET_LINE_REQUEST_TYPE,
                       0, 0, buf, 7, 100);
  if(ret!=7)
  {
    DebugPrint(KERN_INFO "set line reqw failed %d\n",ret);
    return -EIO;
  }
  return 0;
}

int pl2303_write(PRDS_01_CTX pDev,u8 *pBuffer,int BufSize)
{
  int BytesWrite,ret;
  ret=usb_bulk_msg(pDev->udev,usb_sndbulkpipe(pDev->udev,pDev->bOut_Addr),
                      pBuffer,BufSize,&BytesWrite,1000);
  //DebugPrint(KERN_INFO "ret %d write %d bytes\n",ret,BytesWrite);
  if(BytesWrite==0)
  {
   // DebugPrint(KERN_INFO "retr write\n");
    mdelay(50);
    ret=usb_bulk_msg(pDev->udev,usb_sndbulkpipe(pDev->udev,pDev->bOut_Addr),
                      pBuffer,BufSize,&BytesWrite,100);
    //DebugPrint(KERN_INFO "ret %d write %d bytes\n",ret,BytesWrite);                          
  }
  return 0;
}



int pl2303_read(PRDS_01_CTX pDev,u8 *pBuffer,int BufSize)
{
  int BytesRead,ret;
  ret=usb_bulk_msg(pDev->udev,usb_rcvbulkpipe(pDev->udev,pDev->bIn_Addr),
                      pBuffer,BufSize,&BytesRead,1000);
  //DebugPrint(KERN_INFO "ret %d read %d bytes\n",ret,BytesRead);
  if(ret==-EOVERFLOW)
  {
    ret=usb_bulk_msg(pDev->udev,usb_rcvbulkpipe(pDev->udev,pDev->bIn_Addr),
                      pBuffer,BufSize,&BytesRead,1000);
    DebugPrint(KERN_INFO "retr read  ret %d read %d bytes\n",ret,BytesRead);
  }
  if(ret==0)
    return BytesRead;
  return ret;                      
}

int pl2303_set_speed(PRDS_01_CTX pDev,u32 Speed)
{
#if 0
  pl2303_get_line_request(pDev,pDev->pDMACtrlBuf);
#endif
  pDev->pDMACtrlBuf[0]=Speed&0xff;
  pDev->pDMACtrlBuf[1]=(Speed>>8)&0xff;
  pDev->pDMACtrlBuf[2]=(Speed>>16)&0xff;
  pDev->pDMACtrlBuf[3]=(Speed>>24)&0xff;  
  pDev->pDMACtrlBuf[4]=0; //bit cnt
  pDev->pDMACtrlBuf[5]=0; //parity
  pDev->pDMACtrlBuf[6]=8; //bit cnt;
  return pl2303_set_line_request(pDev,pDev->pDMACtrlBuf);
}

int pl2303_init(PRDS_01_CTX pDev)
{
  enum pl2303_type type = type_0;
  if(pDev->udev->descriptor.bDeviceClass == 0x02)
    type = type_1;
  else if (pDev->udev->descriptor.bMaxPacketSize0 == 0x40)
    type = HX;
  else if (pDev->udev->descriptor.bDeviceClass == 0x00)
    type = type_1;
  else if (pDev->udev->descriptor.bDeviceClass == 0xFF)
    type = type_1;
  DebugPrint(KERN_INFO "device type: %d\n", type);  
  pl2303_vendor_read(0x8484, 0, pDev, pDev->pDMACtrlBuf);
  pl2303_vendor_write(0x0404, 0, pDev);
  pl2303_vendor_read(0x8484, 0, pDev, pDev->pDMACtrlBuf);
  pl2303_vendor_read(0x8383, 0, pDev, pDev->pDMACtrlBuf);
  pl2303_vendor_read(0x8484, 0, pDev, pDev->pDMACtrlBuf);
  pl2303_vendor_write(0x0404, 1, pDev);
  pl2303_vendor_read(0x8484, 0, pDev, pDev->pDMACtrlBuf);
  pl2303_vendor_read(0x8383, 0, pDev, pDev->pDMACtrlBuf);
  pl2303_vendor_write(0, 1, pDev);
  pl2303_vendor_write(1, 0, pDev);
  if(type==HX)
    pl2303_vendor_write(2, 0x44, pDev);
  else
    pl2303_vendor_write(2, 0x24, pDev);	
  return 0;
}

int pl2303_open(PRDS_01_CTX pDev)
{
  usb_clear_halt(pDev->udev,usb_sndbulkpipe(pDev->udev,pDev->bOut_Addr));
  usb_clear_halt(pDev->udev,usb_rcvbulkpipe(pDev->udev,pDev->bIn_Addr));
  pl2303_vendor_write(8, 0,pDev);
  pl2303_vendor_write(9, 0,pDev); 
//  pl2303_vendor_write(7,0x3,pDev); 

  return 0;
}
